use std::fmt::{Debug, Display};
use std::ops::{Add, Sub, Mul, Div, Rem};
use std::ops::{AddAssign, SubAssign, MulAssign, DivAssign, RemAssign};
use num::Integer;

#[derive(Debug, PartialEq, Clone)]
pub struct Point<T: Integer> {
    pub x: T,
    pub y: T,
}

impl<T: Integer + Copy> From<T> for Point<T> {
    fn from(value: T) -> Self {
        Self { x: value, y: value }
    }
}

impl<T: Integer> From<(T, T)> for Point<T> {
    fn from(value: (T, T)) -> Self {
        Self { x: value.0, y: value.1 }
    }
}

// Yes, I'm cranky Rust can't allow me to use T: Integer here, I have to define them myself for the types
// Which I cover: i8, u8, i16, u16, i32, u32, i64, u64, i128, and u128
impl TryFrom<String> for Point<i8> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<i8>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<i8>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}

impl TryFrom<String> for Point<i16> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<i16>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<i16>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}

impl TryFrom<String> for Point<i32> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<i32>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<i32>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}

impl TryFrom<String> for Point<i64> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<i64>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<i64>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}

impl TryFrom<String> for Point<i128> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<i128>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<i128>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}

impl TryFrom<String> for Point<u8> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<u8>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<u8>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}

impl TryFrom<String> for Point<u16> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<u16>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<u16>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}

impl TryFrom<String> for Point<u32> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<u32>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<u32>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}

impl TryFrom<String> for Point<u64> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<u64>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<u64>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}

impl TryFrom<String> for Point<u128> {
    type Error = anyhow::Error;
    fn try_from(value: String) -> Result<Self, Self::Error> {
        if !value.contains("(") || !value.contains(")") || !value.contains(",") {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
        let mut work = value.clone().trim().to_string();
        work = work.replace("(", "").replace(")", "").replace(" ", "");
        if let Some(parts) = work.split_once(",") {
            let r = parts.0.parse::<u128>();
            match r {
                Ok(x) => {
                    let r2 = parts.1.parse::<u128>();
                    match r2 {
                        Ok(y) => {
                            return Ok(Point{x, y})
                        }
                        Err(e) => {
                            return Err(anyhow::format_err!("malformed y value, {}", e));
                        }
                    }
                },
                Err(e) => {
                    return Err(anyhow::format_err!("malformed x value, {}", e));
                }
            }
        } else {
            anyhow::bail!("malformed point, expected \"(x, y)\", got {:?}", value);
        }
    }
}

impl<T: Integer + Default> Default for Point<T> {
    fn default() -> Self {
        Self { x: T::default(), y: T::default() }
    }
}

impl<T: Integer + Display> Display for Point<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(format!("({}, {})", self.x, self.y).as_str())
    }
}

impl<T: Integer + Copy> Add<T> for Point<T> {
    type Output = Self;
    fn add(self, rhs: T) -> Self::Output {
        Self::from((self.x + rhs, self.y + rhs))
    }
}

impl<T: Integer> Add<(T, T)> for Point<T> {
    type Output = Self;
    fn add(self, rhs: (T, T)) -> Self::Output {
        Self::from((self.x+rhs.0, self.y+rhs.1))
    }
}

impl<T: Integer + Copy> Sub<T> for Point<T> {
    type Output = Self;
    fn sub(self, rhs: T) -> Self::Output {
        Self::from((self.x-rhs, self.y-rhs))
    }
}

impl<T: Integer> Sub<(T, T)> for Point<T> {
    type Output = Self;
    fn sub(self, rhs: (T, T)) -> Self::Output {
        Self::from((self.x-rhs.0, self.y-rhs.1))
    }
}

impl<T: Integer + Copy> Mul<T> for Point<T> {
    type Output = Self;
    fn mul(self, rhs: T) -> Self::Output {
        Self::from((self.x*rhs, self.y*rhs))
    }
}

impl<T: Integer> Mul<(T, T)> for Point<T> {
    type Output = Self;
    fn mul(self, rhs: (T, T)) -> Self::Output {
        Self::from((self.x*rhs.0, self.y*rhs.1))
    }
}

impl<T: Integer + Copy> Div<T> for Point<T> {
    type Output = Self;
    fn div(self, rhs: T) -> Self::Output {
        Self::from((self.x/rhs, self.y/rhs))
    }
}

impl<T: Integer> Div<(T, T)> for Point<T> {
    type Output = Self;
    fn div(self, rhs: (T, T)) -> Self::Output {
        Self::from((self.x/rhs.0, self.y/rhs.1))
    }
}

impl<T: Integer + Copy> Rem<T> for Point<T> {
    type Output = Self;
    fn rem(self, rhs: T) -> Self::Output {
        Self::from((self.x%rhs, self.y%rhs))
    }
}

impl<T: Integer> Rem<(T, T)> for Point<T> {
    type Output = Self;
    fn rem(self, rhs: (T, T)) -> Self::Output {
        Self::from((self.x%rhs.0, self.y%rhs.1))
    }
}

impl<T: Integer + AddAssign + Copy> AddAssign<T> for Point<T> {
    fn add_assign(&mut self, rhs: T) {
        self.x += rhs;
        self.y += rhs;
    }
}

impl<T: Integer + AddAssign> AddAssign<(T, T)> for Point<T> {
    fn add_assign(&mut self, rhs: (T, T)) {
        self.x += rhs.0;
        self.y += rhs.1;
    }
}

impl<T: Integer + SubAssign + Copy> SubAssign<T> for Point<T> {
    fn sub_assign(&mut self, rhs: T) {
        self.x -= rhs;
        self.y -= rhs;
    }
}

impl<T: Integer + SubAssign> SubAssign<(T, T)> for Point<T> {
    fn sub_assign(&mut self, rhs: (T, T)) {
        self.x -= rhs.0;
        self.y -= rhs.1;
    }
}

impl<T: Integer + MulAssign + Copy> MulAssign<T> for Point<T> {
    fn mul_assign(&mut self, rhs: T) {
        self.x *= rhs;
        self.y *= rhs;
    }
}

impl<T: Integer + MulAssign> MulAssign<(T, T)> for Point<T> {
    fn mul_assign(&mut self, rhs: (T, T)) {
        self.x *= rhs.0;
        self.y *= rhs.1;
    }
}

impl<T: Integer + DivAssign + Copy> DivAssign<T> for Point<T> {
    fn div_assign(&mut self, rhs: T) {
        self.x /= rhs;
        self.y /= rhs;
    }
}

impl<T: Integer + DivAssign> DivAssign<(T, T)> for Point<T> {
    fn div_assign(&mut self, rhs: (T, T)) {
        self.x /= rhs.0;
        self.y /= rhs.1;
    }
}

impl<T: Integer + RemAssign + Copy> RemAssign<T> for Point<T> {
    fn rem_assign(&mut self, rhs: T) {
        self.x %= rhs;
        self.y %= rhs;
    }
}

impl<T: Integer + RemAssign> RemAssign<(T, T)> for Point<T> {
    fn rem_assign(&mut self, rhs: (T, T)) {
        self.x %= rhs.0;
        self.y %= rhs.1;
    }
}
